#!/usr/bin/env bash
# Copyright 2018 The Bazel Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

set -o pipefail -o errexit -o nounset

# --- begin runfiles.bash initialization v2 ---
# Copy-pasted from the Bazel Bash runfiles library v2.
f=bazel_tools/tools/bash/runfiles/runfiles.bash
source "${RUNFILES_DIR:-/dev/null}/$f" 2>/dev/null || \
  source "$(grep -sm1 "^$f " "${RUNFILES_MANIFEST_FILE:-/dev/null}" | cut -f2- -d' ')" 2>/dev/null || \
  source "$0.runfiles/$f" 2>/dev/null || \
  source "$(grep -sm1 "^$f " "$0.runfiles_manifest" | cut -f2- -d' ')" 2>/dev/null || \
  source "$(grep -sm1 "^$f " "$0.exe.runfiles_manifest" | cut -f2- -d' ')" 2>/dev/null || \
  { echo>&2 "ERROR: cannot find $f"; exit 1; }; f=;
# --- end runfiles.bash initialization v2 ---

case "$(uname -s)" in
  CYGWIN*|MINGW32*|MSYS*|MINGW*)
    ON_WINDOWS=1
    ;;
  *)
    ON_WINDOWS=0
    ;;
esac

function error() {
  echo >&2 "installer_test.sh: $@"
  exit 1
}

# installer rule outut
declare -r INSTALLER="$(rlocation $TEST_WORKSPACE/$1)"
# srcs attribute of installer rule
declare -r FILE="$(rlocation $TEST_WORKSPACE/$2)"

while (( "$#" >  2 )); do
  case "$3" in
    -subdir=*)
      # target_subdir attribute of installer rule
      declare -r SUBDIR="${3#-subdir=}"
    ;;
    -executable=*)
      # executable attribute of installer rule
      declare -r EXECUTABLE="${3#-executable=}"
    ;;
    *)
      error "unknown flag $3"
    ;;
  esac
  shift
done

if [[ -z "${SUBDIR:-}" ]]; then
  declare -r SUBDIR=""
fi

function setup() {
  [[ -r "${FILE}" ]] ||
    error "Can't read ${FILE}"
  [[ -x "${INSTALLER}" ]] ||
    error "Can't run ${INSTALLER}"
  mktemp --directory --tmpdir="${TEST_TMPDIR}"
}

function verify_install() {
  local expected_path="$1"
  local source_path="$2"

  if [[ ! -f "${expected_path}" ]]; then
    error "expected file ${expected_path} doesn't exist"
  fi

  local mode
  mode="$(stat --dereference --format=%a "${expected_path}")"
  local expected_mode
  if [[ "${EXECUTABLE}" = "True" ]]; then
    expected_mode="755"
  else
    expected_mode="644"
  fi

  # Do not check permission on Windows
  if [[ "${ON_WINDOWS}" -eq 0 && "${mode}" != "${expected_mode}" ]]; then
    echo >&2 "mode of file ${expected_path} (${mode}) do not match ${expected_mode}"
    return 1
  fi

  if ! diff -q "${source_path}" "${expected_path}"; then
    echo >&2 "contents of file ${expected_path} do not match ${source_path}"
    return 1
  fi

}

function test_basic_install() {
  local tmpdir="$1"

  if ! "${INSTALLER}" "${tmpdir}"; then
    echo >&2 "installer didn't exit with code 0"
    return 1
  fi

  local expected_path
  if [[ -n "${SUBDIR}" ]]; then
    expected_path="${tmpdir}/${SUBDIR}/$(basename "${FILE}")"
  else
    expected_path="${tmpdir}/$(basename "${FILE}")"
  fi

  verify_install "${expected_path}" "${FILE}"
}

function test_prefix_install() {
  local tmpdir="$1"

  if ! "${INSTALLER}" "${tmpdir}/a/b/c"; then
    echo >&2 "installer didn't exit with code 0"
    return 1
  fi

  local expected_path
  if [[ -n "${SUBDIR}" ]]; then
    expected_path="${tmpdir}/a/b/c/${SUBDIR}/$(basename "${FILE}")"
  else
    expected_path="${tmpdir}/a/b/c/$(basename "${FILE}")"
  fi

  verify_install "${expected_path}" "${FILE}"
}

function test_generated_warning() {
  # On Windows, ${INSTALLER} is a .exe and not the script, so
  # skip this test
  if [[ "${ON_WINDOWS}" -eq 0 ]]; then
    if ! grep -q "DO NOT EDIT THIS FILE" "${INSTALLER}"; then
      echo >&2 "installer $INSTALLER doesn't have a DO NOT EDIT warning"
      exit 1
    fi
    if grep "@@.*@@" "${INSTALLER}" >&2; then
      echo >&2 "installer seems to contain unassigned template variables"
      exit 1
    fi
  fi
}

function main() {
  local tmpdir
  local test
  for test in test_basic_install test_prefix_install test_generated_warning; do
    tmpdir="$(setup)"
    if "${test}" "${tmpdir}"; then
      echo "Passed ${test}"
    else
      echo "Failed ${test}"
      exit 1
    fi
  done
}

main
